from ._scene_internal import Line, Mesh, Point


class Scene:
    """
    Simple scenegraph for quickly debugging 3D meshes, polylines, and points.

    Args:
        point_radius (float): The desired radius of the markers that are
            rendered for points. The default is `1.0`.
    """

    def __init__(self, point_radius=1.0):
        self.point_radius = point_radius
        self.children = []

    def add_meshes(self, *meshes):
        """
        Add one or more meshes to the scene.

        Args:
            meshes (list): One or more instances of `lacecore.Mesh`.

        Return:
            Self for chaining.

        See also:
            https://lacecore.readthedocs.io/en/latest/
        """
        for mesh in meshes:
            self.children.append(Mesh(mesh=mesh))
        return self

    def add_lines(self, *polylines, color="red"):
        """
        Add one or more polylines to the scene.

        Args:
            polylines (list): One or more instances of `polliwog.Polyline`.
            color (object): A CSS named color or an tuple of RGB values,
                scaled from 0 to 1.

        Return:
            Self for chaining.

        See also:
            https://polliwog.readthedocs.io/en/latest/
        """
        for polyline in polylines:
            self.children.append(Line(polyline=polyline, color=color))
        return self

    def add_points(self, *points, name=None, color="red"):
        """
        Add one or more points to the scene.

        Args:
            points (list): One or more `(3,)` NumPy arrays or a `kx3` array.
            name (str): The name of a point to add. Do not use this when providing
                more than one point.
            color (object): A CSS named color or an tuple of RGB values,
                scaled from 0 to 1.

        Return:
            Self for chaining.
        """
        if len(points) > 0 and name is not None:
            raise ValueError(
                "When more than one point is provided, expected `name` to be None"
            )
        for point in points:
            self.children.append(Point(point, name=name, color=color))
        return self

    def write(self, filename):
        """
        Write a COLLADA file for this scene.

        Args:
            filename (str): The filename to write (e.g. `debug.dae`).
        """
        import os
        from ._collada import collada_from_scene

        _, extension = os.path.splitext(filename)
        if extension != ".dae":
            raise ValueError("Expected filename to end with .dae")

        dae = collada_from_scene(self)
        dae.write(filename)



from ._color import normalize_color


class Object:
    pass


class Mesh(Object):
    def __init__(self, mesh, name=None):
        self.mesh = mesh
        self.name = name


class Line(Object):
    def __init__(self, polyline, color):
        # Validate.
        normalize_color(color)

        self.polyline = polyline
        self.color = color


class Point(Object):
    def __init__(self, point, color, name=None):
        # Validate.
        normalize_color(color)

        self.point = point
        self.name = name
        self.color = color



from collada import Collada
from collada.geometry import Geometry
from collada.material import Effect, Material
from collada.scene import GeometryNode, MaterialNode, Node, Scene
from collada.source import FloatSource, InputList
import numpy as np
from polliwog import Polyline
from toolz import groupby
from vg.compat import v1 as vg
from ._color import normalize_color
from ._scene_internal import (
    Line as InternalLine,
    Mesh as InternalMesh,
    Point as InternalPoint,
)


def create_material(collada, name, color=(1, 1, 1)):
    effect = Effect(
        f"{name}_effect",
        [],
        "lambert",
        diffuse=color,
        specular=(0, 0, 0),
        double_sided=True,
    )
    material = Material(name, name, effect)
    collada.effects.append(effect)
    collada.materials.append(material)
    return MaterialNode(name, material, inputs=[])


def geometry_node_from_mesh(collada, mesh, name):
    vertex_source_name = f"{name}_verts"
    geometry = Geometry(
        collada,
        name or "geometry0",
        str(mesh),
        [FloatSource(vertex_source_name, mesh.v.ravel(), ("X", "Y", "Z"))],
    )
    input_list = InputList()
    input_list.addInput(0, "VERTEX", f"#{vertex_source_name}")

    material_name = f"{name}_material"
    triset = geometry.createTriangleSet(mesh.f.ravel(), input_list, material_name)

    geometry.primitives.append(triset)
    collada.geometries.append(geometry)

    return GeometryNode(geometry, [create_material(collada, name=material_name)])


def geometry_node_from_segments(collada, vertices, edges, color, name, description):
    vg.shape.check(locals(), "vertices", (-1, 3))
    vg.shape.check(locals(), "edges", (-1, 2))

    vertex_source_name = f"{name}_verts"
    vertex_source = FloatSource(vertex_source_name, vertices, ("X", "Y", "Z"))
    geometry = Geometry(collada, name, description, [vertex_source])
    input_list = InputList()
    input_list.addInput(0, "VERTEX", f"#{vertex_source_name}")

    material_name = f"{name}_material"
    lineset = geometry.createLineSet(edges.ravel(), input_list, material_name)

    geometry.primitives.append(lineset)
    collada.geometries.append(geometry)

    return GeometryNode(
        geometry, [create_material(collada, name=material_name, color=color)]
    )


def geometry_node_from_polyline(collada, polyline, color, name):
    if not isinstance(polyline, Polyline):
        raise ValueError("Expected a Polyline")
    return geometry_node_from_segments(
        collada=collada,
        vertices=np.ascontiguousarray(polyline.v),
        edges=polyline.e,
        color=color,
        name=name,
        description=str(polyline),
    )


def geometry_node_from_points(collada, points, radius, color, name):
    vg.shape.check(locals(), "points", (-1, 3))

    offset = radius * np.eye(3)
    segments = np.repeat(points, 6, axis=0).reshape(-1, 3, 2, 3)
    segments[:, :, 0] = segments[:, :, 0] + offset
    segments[:, :, 1] = segments[:, :, 1] - offset

    vertices = segments.reshape(-1, 3)
    edges = np.arange(len(vertices)).reshape(-1, 2)

    return geometry_node_from_segments(
        collada=collada,
        vertices=vertices,
        edges=edges,
        color=color,
        name=name,
        description=f"{len(points)} points",
    )


def collada_from_scene(scene, name="triagain"):
    collada = Collada()

    geometry_nodes = (
        [
            geometry_node_from_mesh(
                collada=collada,
                mesh=child.mesh,
                name=f"mesh_geometry_{i}",
            )
            for i, child in enumerate(
                child for child in scene.children if isinstance(child, InternalMesh)
            )
        ]
        + [
            geometry_node_from_polyline(
                collada,
                child.polyline,
                name=f"polyline_geometry_{i}",
                color=normalize_color(child.color),
            )
            for i, child in enumerate(
                child for child in scene.children if isinstance(child, InternalLine)
            )
        ]
        + [
            geometry_node_from_points(
                collada,
                points=np.array([point.point for point in points]),
                radius=scene.point_radius,
                color=normalize_color(color),
                name=f"point_geometry_{i}",
            )
            for i, (color, points) in enumerate(
                groupby(
                    lambda point: point.color,
                    [
                        child
                        for child in scene.children
                        if isinstance(child, InternalPoint)
                    ],
                ).items()
            )
        ]
    )

    scene = Scene(name, [Node("root", children=geometry_nodes)])
    collada.scenes.append(scene)
    collada.scene = scene
    return collada



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tri_again']

package_data = \
{'': ['*']}

install_requires = \
['numpy',
 'polliwog>=2.0.0,<3',
 'pycollada>=0.7.1,<0.8',
 'toolz>=0.10.0,<0.12.0',
 'vg>=2.0.0',
 'webcolors>=1.11.1,<2']

setup_kwargs = {
    'name': 'tri-again',
    'version': '1.1.1',
    'description': 'Scenegraph for quickly debugging 3D meshes, polylines, and points',
    'long_description': None,
    'author': 'Paul Melnikow',
    'author_email': 'github@paulmelnikow.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/lace/tri-again',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<3.9',
}


setup(**setup_kwargs)



from webcolors import name_to_rgb_percent


def normalize_color(color):
    if isinstance(color, tuple):
        if len(color) != 3 or min(v for v in color) < 0 or max(v for v in color) > 1:
            raise ValueError(
                "When specified as a tuple, color should be three floats between 0 and 1"
            )
        return color
    else:
        return tuple(0.01 * float(x.strip("%")) for x in name_to_rgb_percent(color))
